# program to read and manipulate USHCN data
# input data is assumed to be csv

# imports

from __future__ import division
import math
import numpy as np
import numpy.ma as ma
import scipy as sp
import scipy.stats as ss
import scipy.stats.mstats as ms
import csv

# constants, declarations

# input data files
datafilenames = ( 'Watts/9641C_201008_raw_avg.csv',
                 'Watts/9641C_201008_tob_avg.csv',
                 'Watts/9641C_201008_F52_avg.csv',
                 'Watts/9641C_201008_raw_max.csv',
                 'Watts/9641C_201008_tob_max.csv',
                 'Watts/9641C_201008_F52_max.csv',
                 'Watts/9641C_201008_raw_min.csv',
                 'Watts/9641C_201008_tob_min.csv',
                 'Watts/9641C_201008_F52_min.csv' )
NDATA = 9

# input station file
stnfilename = 'Watts/RATINGS.csv'
#stnfilename = 'Watts/RATINGSproxy5.csv'
#stnfilename = 'Watts/RATINGSproxy12_1real12_2.csv'

# output data file
outfilename = 'Watts/output01'

# maximum number of stations
MAXSTN = 1500

# good station ratings
good = [ "1", "2", "3" ]
MAXGOOD = 1000

# bad station ratings
bad = [ "4", "5" ]
MAXBAD = 1000

# first year to use
#FIRSTYR = 1895
FIRSTYR = 1979

# last year to use
#LASTYR = 2009
LASTYR = 2008

# subtract off monthly normals?
#donorm = "yes"
donorm = "no"

# first year of normals
#FNORM = 1951
FNORM = 1979

# last year of normals
#LNORM = 1980
LNORM = 2008

# first year of trend
#TRST = 1895
TRST = 1979

# last year of trend
#TREND = 2009
TREND = 2008

# number of full monte carlo runs
NMONTE = 10000
#NMONTE = 1000
#NMONTE = 0

# list of region numbers
regionlist = ('1','2','3','4','5','6','7','8','9')

# size of climate regions
regionsizes = np.array([.06021, .08428, .10271, .09715, .15551, .18822, .14053, .08230, .08908])

# Use NARR?
usenarr = "yes"
narrdir = "Watts/NARR/CRN125/"
#####################################

def fullyear (inputrow):
   yr = np.zeros((12))
   for i in range(0,12):
      yr[i] = inputrow[3+2*i]
   return(yr)

#####################################

# create some variables

goodstn = []
badstn = []
goodreg = []
badreg = []
setstn = []
setname = []
setstate = []
setreg = []
setstat = []

nyr = LASTYR+1-FIRSTYR

years = range(FIRSTYR, LASTYR+1)
setdata = np.zeros((NDATA,MAXGOOD+MAXBAD,nyr,12)) - 999.9
mgoodata = ma.zeros((NDATA,MAXGOOD,nyr,12)) - 999.9
mbadata = ma.zeros((NDATA,MAXBAD,nyr,12)) - 999.9

# find the good and the bad

ngood = 0
nbad = 0
nset = 0
stnfile = open(stnfilename, "rU")
stnrows = csv.reader(stnfile)
for row in stnrows:
   if row[5] in good:
      goodstn += [ row[0] ]
      goodreg += [ row[6] ]
      setstn += [ row[0] ]
      setname += [ row[1] ]
      setstate += [ row[2] ]
      setreg += [ row[6] ]
      setstat += [ 1 ]
      ngood +=1
      nset +=1
   if row[5] in bad:
      badstn += [ row[0] ]
      badreg += [ row[6] ]
      setstn += [ row[0] ]
      setname += [ row[1] ]
      setstate += [ row[2] ]
      setreg += [ row[6] ]
      setstat += [ 0 ]
      nbad +=1
      nset +=1
   
stnfile.close()   
print "finished with station ID file"
asetreg=np.array(setreg)

# read in the data

idat = -1
for datafilename in datafilenames:
   idat += 1
   datafile = open(datafilename, "rU")
   datarows = csv.reader(datafile)
   for row in datarows:
      if row[0] in setstn:
         if int(row[2]) in years:
            setdata[idat,setstn.index(row[0]),years.index(int(row[2])),:] = fullyear(row)/10.
   datafile.close()
print "Finished with raw data file"

# set missing data flags, convert to Celsius
msetdata = ma.masked_equal(setdata,-999.9)
msetdata = (msetdata - 32.) * 5./9.

# Ingest NARR data and compute differences
if usenarr == "yes":
   narrdata = np.zeros((MAXGOOD+MAXBAD,nyr*12))
   for istn in range(0, nset):
      narrfile = open(narrdir+setname[istn]+" "+setstate[istn]+".csv", "rU")
      narrows = csv.reader(narrfile)
      imo = 0
      for row in narrows:
         narrdata[istn,imo] = float(row[0]) - 273.15
         imo += 1
      if imo != 12*(LASTYR-FIRSTYR+1):
         print "That is strange imo is",imo,12*(LASTYR-FIRSTYR+1)
      narrfile.close()

   narrdata = np.reshape(narrdata,(MAXGOOD+MAXBAD,nyr,12))
   msetdata = msetdata - narrdata
   print "Done ingesting NARR"

# compute monthly climatology for normal period
mav = ma.average(msetdata[:,:,FNORM-FIRSTYR:LNORM+1-FIRSTYR,:],axis=2)

if donorm == "yes":
# compute anomalies relative to monthly normals
   mav = ma.expand_dims(mav,axis=2)

#Hack number 1: use averages
#   msetdata = msetdata - mav
   msetdata = mav
   print "Done computing anomalies"

# fill missing values?

#************************************************************************************
slopes = np.zeros((NDATA,3,10,NMONTE+1))
dtrslopes = np.zeros((3,3,10,NMONTE+1))
xdata=np.arange(TRST,TREND+1)
ntests = 0
itests = 0
fracgood = ngood/(ngood+nbad)

while ntests < NMONTE + 1:
# load good and bad data arrays
   for stn in setstn:
      if setstat[setstn.index(stn)] == 1:
         mgoodata[:,goodstn.index(stn),:,:] = msetdata[:,setstn.index(stn),:,:]
      else:
         if setstat[setstn.index(stn)] == 0:
            mbadata[:,badstn.index(stn),:,:] = msetdata[:,setstn.index(stn),:,:]
         else:
            print "Station ",stn, " slipped through the cracks."

# compute regional averages, good and bad, by averaging by month and then by year
   regav=np.zeros((NDATA,2,10))
   for ireg in regionlist:
      weights = np.zeros((MAXGOOD))
      for istn in range(0,ngood):
         if goodreg[istn]==ireg:
            weights[istn]=1
      regav[:,0,int(ireg)] = ma.average(ma.average(mgoodata[:,:,0,:],axis=1,weights=weights)[:,:],axis=1)
      weights = np.zeros((MAXBAD))
      for istn in range(0,nbad):
         if badreg[istn]==ireg:
            weights[istn]=1
      regav[:,1,int(ireg)] = ma.average(ma.average(mbadata[:,:,0,:],axis=1,weights=weights)[:,:],axis=1)

   regav[:,:,0] = ma.average(regav[:,:,1:10],axis=2,weights=regionsizes)

   for idat in range(0,NDATA):
      for i in range(0,10):

         slopes[idat,0,i,ntests] = regav[idat,0,i]
         if ntests == 0:
            print "Dataset",idat," Region",i," good siting nardif = ",regav[idat,0,i]
         slopes[idat,1,i,ntests] = regav[idat,1,i] 
         if ntests == 0:
            print "Dataset",idat," Region",i," bad siting nardif = ",regav[idat,1,i]
         slopes[idat,2,i,ntests] = regav[idat,1,i]-regav[idat,0,i]
         if ntests == 0:
            print "Dataset",idat," Region",i," difference nardif = ",regav[idat,1,i]-regav[idat,0,i]

   for idat in range(0,3):
      for i in range(0,10):
          dtrslopes[idat,0,i,ntests] = regav[idat+3,0,i]-regav[idat+6,0,i]
          dtrslopes[idat,1,i,ntests] = regav[idat+3,1,i]-regav[idat+6,1,i]
          dtrslopes[idat,2,i,ntests] = regav[idat+3,1,i]-regav[idat+6,1,i]-regav[idat+3,0,i]+regav[idat+6,0,i]

   ntests +=1
   itests +=1
   if itests == 100:
      print "Finished iteration ",ntests
      itests = 0

# Randomize and test the good/bad classification
   goodrandom = 0
   while goodrandom == 0:
      setstat = np.random.permutation(setstat[0:nset]) 
      goodrandom = 1
      for ireg in regionlist:
         regtotal = np.sum( np.where(asetreg==ireg,1,0) )
         reggood = np.sum( np.where(asetreg==ireg,setstat,0) )
         if reggood < 2:
            goodrannndom = 0
         if regtotal-reggood < 2:
            goodrandom = 0

   goodstn = []
   goodreg = []
   badstn = []
   badreg = []
   for istn in range(0,nset):
      if setstat[istn] == 1:
         goodstn += [ setstn[istn] ]
         goodreg += [ setreg[istn] ]
      else:
         badstn += [ setstn[istn] ]
         badreg += [ setreg[istn] ]

#**************************************End of Monte Carlo iteration*********************************

sortslopes = np.sort(slopes[:,:,:,1:NMONTE+1],axis=3)
sortdtrslopes = np.sort(dtrslopes[:,:,:,1:NMONTE+1],axis=3)

for id in range(0,NDATA):
   for ireg in range(0,10):
      spot1=np.searchsorted(sortslopes[id,2,ireg,:],slopes[id,2,ireg,0])
      spot2=np.searchsorted(sortslopes[id,2,ireg,:],slopes[id,2,ireg,0]+.000001)
      spot=(spot1+spot2)/2.
      if(spot>NMONTE/2.):
         pvalue=2.*(NMONTE-spot)/NMONTE
      else:
         pvalue=2.*spot/NMONTE      
      print datafilenames[id]," Region = ", ireg, slopes[id,:,ireg,0], pvalue

for id in range(0,3):
   for ireg in range(0,10):
      spot1=np.searchsorted(sortdtrslopes[id,2,ireg,:],dtrslopes[id,2,ireg,0])
      spot2=np.searchsorted(sortdtrslopes[id,2,ireg,:],dtrslopes[id,2,ireg,0]+.000001)
      spot=(spot1+spot2)/2.
      if(spot>NMONTE/2.):
         pvalue=2.*(NMONTE-spot)/NMONTE
      else:
         pvalue=2.*spot/NMONTE
      print datafilenames[id]," Region = ", ireg, " dtr ", dtrslopes[id,:,ireg,0], pvalue
